# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
Simple lexing using python regular expressions.
'''

import re
from typing import Container, Dict, FrozenSet, Iterable, Iterator, List, NamedTuple, Optional, Pattern, Tuple, Union, cast

from tolkien import Source, Token

from .string import iter_str


class LexError(Exception): pass


class LexMode:
  def __init__(self, name:str, kinds:Iterable[str], *, indents=False) -> None:
    self.name = name
    self.kinds = list(iter_str(kinds))
    self.indents = indents
    self.kind_set:FrozenSet[str] = frozenset() # Filled in by Lexer.
    self.regex:Pattern = cast(Pattern, None) # Filled in by Lexer.

  def __repr__(self) -> str:
    return f'{type(self).__name__}({self.name!r}, kinds={self.kinds}, kind_set={self.kind_set}, indents={self.indents})'


class KindPair(NamedTuple):
  'A pair of kinds that, when encountered in sequence, trigger a mode transition.'
  prev:str
  curr:str

  @staticmethod
  def mk(kind:'LexTransKind') -> 'KindPair':
    if isinstance(kind, str): return KindPair(prev='', curr=kind)
    if isinstance(kind, KindPair): return kind
    raise Lexer.DefinitionError(f'expected `str` or `KindPair`; received {kind!r}')


LexTransKind = Union[str,KindPair]
LexTransKinds = Union[LexTransKind,Iterable[LexTransKind]]
LexTransTuple = Tuple[KindPair,...]

class LexTrans:
  def __init__(self, base:Iterable[str], *, kind:LexTransKinds, mode:str, pop:LexTransKinds, consume:bool) -> None:
    self.bases = tuple(iter_str(base)) # Parent modes.
    self.kinds:LexTransTuple = _mk_pairs_tuple(kind) # Parent mode token kind or pair that causes push.
    self.mode = mode # Child mode.
    self.pops:LexTransTuple = _mk_pairs_tuple(pop)
    self.pop_dict = _mk_trans_pop_dict(self.pops) # Child mode token kinds or pairs as (curr -> prev or '') mapping.
    self.consume = consume # If false, base will consider this token for further push/pop mode transitions.

  def __repr__(self) -> str:
    return f'{type(self).__name__}({self.bases!r}, kinds={self.kinds}, mode={self.mode!r}, pops={self.pops}, consume={self.consume})'

  def should_pop(self, frame_token:Token, token:Token, prev_kind:str) -> bool:
    try: pop_prev = self.pop_dict[token.kind]
    except KeyError: return False
    return not pop_prev or prev_kind == pop_prev
    # TODO: support exact token content matching.


def _mk_pairs_tuple(pop:LexTransKinds) -> LexTransTuple:
  if isinstance(pop, (str, KindPair)): pop = (pop,)
  return tuple(KindPair.mk(el) for el in pop)


def _mk_trans_pop_dict(pops:LexTransTuple) -> Dict[str,str]:
  d:Dict[str,str] = {}
  for pair in pops:
    if pair.curr in d:
      raise Lexer.DefinitionError(f'repeated pop kind key: {pair.curr!r}')
    d[pair.curr] = pair.prev
  return d



class Lexer:
  '''
  Define a Lexer using python regular expressions.
  This class is used to prototype (and bootstrap) Legs lexers.
  Note: A zero-length match, e.g. r'^' causes an exception; otherwise the stream would never advance.
  '''

  class DefinitionError(Exception): pass

  def __init__(self, *, flags='', patterns:Dict[str,str], modes:Iterable[LexMode]=(),
   transitions:Iterable[LexTrans]=()) -> None:

    # Validate flags.
    for flag in flags:
      if flag not in 'aiLmsux':
        raise Lexer.DefinitionError(f'invalid global regex flag: {flag}')
    flags_pattern = f'(?{flags})' if flags else ''
    is_extended = 'x' in flags

    # Validate patterns.
    if not patterns: raise Lexer.DefinitionError('Lexer instance must define at least one pattern')
    self.patterns: Dict[str,str] = {}
    for n, v in patterns.items():
      validate_name(n)
      if n == 'invalid':
        raise Lexer.DefinitionError(f'{n!r} pattern name collides with the invalid token name')
      if not isinstance(v, str): # TODO: also support bytes.
        raise Lexer.DefinitionError(f'{n!r} pattern value must be a string; found {v!r}')
      pattern = f'{flags_pattern}(?P<{n}>{v})'
      try: r = re.compile(pattern) # compile each expression by itself to improve error clarity.
      except Exception as e:
        lines = [f'{n!r} pattern is invalid: {pattern}']
        if flags: lines.append(f'global flags: {flags!r}')
        if is_extended and re.search('(?<!\\)#)', v): lines.append('unescaped verbose mode comments break lexer')
        msg = '\n  note: '.join(lines)
        raise Lexer.DefinitionError(msg) from e
      for group_name in r.groupindex:
        if group_name in patterns and group_name != n:
          raise Lexer.DefinitionError(f'{n!r} pattern contains a conflicting capture group name: {group_name!r}')
      self.patterns[n] = pattern

    # Validate and compile modes.
    if not modes:
      modes = [LexMode('main', kinds=self.patterns)]
    self.modes: Dict[str,LexMode] = {}
    main = None
    for mode in modes:
      if not isinstance(mode, LexMode): raise Lexer.DefinitionError(f'expected LexMode; received: {mode!r}')
      if not main: main = mode.name
      if mode.name in self.modes:
        raise Lexer.DefinitionError(f'duplicate mode name: {mode.name!r}')
      if mode.name in patterns:
        raise Lexer.DefinitionError(f'mode name conflicts with pattern name: {mode.name!r}')
      kind_set = set()
      for kind in iter_str(mode.kinds):
        if re.fullmatch(r'\w+', kind): # Single name.
          if kind in patterns: kind_set.add(kind)
          else: raise Lexer.DefinitionError(f'mode {mode.name!r} includes nonexistent pattern: {kind!r}')
        else: # Pattern.
          try:
            matching = {p for p in self.patterns if re.fullmatch(kind, p)}
          except Exception:
            raise Lexer.DefinitionError(f'mode {mode.name!r} includes invalid pattern regex: {kind}')
          if not matching:
            raise Lexer.DefinitionError(f'mode {mode.name!r} wildcard pattern regex does not match any patterns: {kind}')
          kind_set.update(matching)
      self.modes[mode.name] = mode
      mode.kind_set = frozenset(kind_set)
      choice_sep = '\n| ' if 'x' in flags else '|'
      mode.regex = re.compile(choice_sep.join(pattern for name, pattern in self.patterns.items() if name in kind_set))
      #^ note: iterate over self.patterns.items (not pattern_names) because the dict preserves the original pattern order.

    kinds = list(self.patterns)
    if any(mode.indents for mode in modes):
      kinds.extend(['indent', 'dedent'])
    self.kinds = frozenset(kinds)

    # Validate transitions.
    assert main is not None
    self.main:str = main
    self.transitions:Dict[Tuple[str,str],Tuple[LexTrans,str]] = {}
    for trans in transitions:
      if not isinstance(trans, LexTrans): raise Lexer.DefinitionError(f'expected `LexTrans`; received {trans!r}')
      for base in trans.bases:
        if base not in self.modes: raise Lexer.DefinitionError(f'unknown parent mode: {base!r}')
      if trans.mode not in self.modes: raise Lexer.DefinitionError(f'unknown child mode: {trans.mode!r}')
      for label, pairs in [('push', trans.kinds), ('pop', trans.pops)]:
        for pair in pairs:
          for i, k in enumerate(pair):
            if i == 0 and not k: continue # Prev is allowed to be empty.
            if k and k not in self.kinds: raise Lexer.DefinitionError(f'unknown mode {label} pattern: {k!r}')
      for base in trans.bases:
        for kind_pair in trans.kinds:
          key = (base, kind_pair.curr)
          try: _, existing = self.transitions[key]
          except KeyError: # No conflict.
            self.transitions[key] = (trans, kind_pair.prev)
          else: # Conflict.
            raise Lexer.DefinitionError(f'conflicting transitions:\n  {existing}\n  {trans}')


  def _lex_inv(self, pos:int, end:int, mode:str) -> Token:
    return Token(pos=pos, end=end, mode=mode, kind='invalid')


  def _lex_one(self, regex:Pattern, source:Source[str], pos:int, end:int, mode:str) -> Token:
    m = regex.search(source.text, pos)
    if not m:
      return self._lex_inv(pos=pos, end=end, mode=mode)
    p, e = m.span()
    if pos < p:
      return self._lex_inv(pos=pos, end=p, mode=mode)
    if p == e:
      raise Lexer.DefinitionError(f'Zero-length patterns are disallowed.\n  kind: {m.lastgroup}; match: {m}')
    kind = m.lastgroup
    assert isinstance(kind, str)
    return Token(pos=p, end=e, mode=mode, kind=kind)


  def _lex(self, stack:List[Tuple[LexTrans,Token]], source:Source[str], pos:int, end:int, drop:Container[str], eot:bool
   ) -> Iterator[Token]:
    assert isinstance(source, Source)
    prev_kind = ''
    indent_stack:List[int] = []
    while pos < end:
      # Get the current frame and mode.
      frame_trans, frame_token = stack[-1]
      mode = self.modes[frame_trans.mode]
      # Lex one token.
      token = self._lex_one(mode.regex, source, pos=pos, end=end, mode=mode.name)
      kind = token.kind
      pos = token.end
      # Check if we should synthesize indent or dedent tokens.
      if mode.indents and prev_kind == 'newline':
        if kind == 'spaces': # Generate indent or dedent tokens.
          indent_len = len(source[token])
          while indent_stack and indent_stack[-1] > indent_len:
            yield token.pos_token(kind='dedent')
            prev_kind = 'dedent' # Hack to allow for transitions on spaces following indent/dedent.
            indent_stack.pop()
          if not indent_stack or indent_stack[-1] < indent_len:
            yield token.pos_token(kind='indent')
            prev_kind = 'indent' # Hack to allow for transitions on spaces following indent/dedent.
            indent_stack.append(indent_len)
        elif kind != 'newline': # Empty lines do not affect indent stack. All others pop the entire indent stack.
          while indent_stack:
            yield token.pos_token(kind='dedent')
            prev_kind = 'dedent' # Hack to allow for transitions on spaces following indent/dedent.
            indent_stack.pop()
      # Yield the current token.
      if kind not in drop:
        yield token
      # Perform mode transitions.
      try:
        # Check if we should pop one or more modes.
        while frame_trans.should_pop(frame_token=frame_token, token=token, prev_kind=prev_kind):
          stack.pop()
          if frame_trans.consume:
            raise _BreakFromModeSwitching # This frame "consumes" the token.
          frame_trans, frame_token = stack[-1]
        # Check if we should push a child mode.
        try: push_trans, push_prev_kind = self.transitions[(frame_trans.mode, kind)]
        except KeyError: pass
        else:
          if not push_prev_kind or prev_kind == push_prev_kind:
            stack.append((push_trans, token))
            kind = '' # Make prev_kind empty on the next iteration.
      except _BreakFromModeSwitching: pass
      prev_kind = kind
    while indent_stack:
      yield token.pos_token(kind='dedent')
      prev_kind = 'dedent' # Hack to allow for transitions on spaces following indent/dedent.
      indent_stack.pop()
    if eot:
      yield eot_token(source, mode=stack[-1][0].mode)


  def lex(self, source:Source[str], pos:int=0, end:Optional[int]=None, drop:Container[str]=(), eot=False) -> Iterator[Token]:
    if not isinstance(source, Source): raise TypeError(source)
    text = source.text
    if pos < 0:
      pos = len(text) + pos
    if end is None:
      end = len(text)
    elif end < 0:
      end = len(text) + end
    _e:int = end # typing hack.
    return self._lex(stack=[self.root_frame(mode=self.main)], source=source, pos=pos, end=_e, drop=drop, eot=eot)


  def lex_stream(self, *, name:str, stream:Iterable[str], drop:Container[str]=(), eot=False) -> Iterator[Tuple[Source[str], Token]]:
    '''
    Note: the yielded Token objects have positions relative to input string that each was lexed from.
    TODO: fix the line numbers.
    '''
    stack = [self.root_frame(mode=self.main)]
    source = Source(name=name, text='')
    for text in stream:
      if text:
        source = Source(name=name, text=text)
        for token in self._lex(stack=stack, source=source, pos=0, end=len(source.text), drop=drop, eot=False):
          yield (source, token)
    if eot:
      yield (source, eot_token(source, mode=stack[-1][0].mode))


  def root_frame(self, mode:str) -> Tuple[LexTrans,Token]:
    'The root frame is degenerate; because it has an empty pops set it will never get popped; base and kind are never accessed.'
    return (LexTrans(base='', kind='', mode=mode, pop=frozenset(), consume=False), Token(0, 0, mode='', kind=''))


class _BreakFromModeSwitching(Exception): pass


def eot_token(source:Source[str], mode:str) -> Token:
  'Create a token representing the end-of-text.'
  end = len(source.text)
  return Token(pos=end, end=end, mode=mode, kind='end_of_text')


def validate_name(name:str) -> str:
  if not valid_name_re.fullmatch(name):
    raise Lexer.DefinitionError(f'invalid name: {name!r}')
  if name in reserved_names:
    raise Lexer.DefinitionError(f'name is reserved: {name!r}')
  return name


valid_name_re = re.compile(r'[A-Za-z_]\w*')

reserved_names = frozenset({'end_of_text'})


whitespace_patterns = dict(
  indents   = r'(?m:^\ +)',
  spaces    = r'\ +',
  ind_tabs  = r'(?m:^\ +)',
  tabs      = r'\t+',
  newline   = r'\n',
)


c_like_punctuation_patterns = dict(

  paren_o     = r'\(',
  paren_c     = r'\)',
  brack_o     = r'\[',
  brack_c     = r'\]',
  brace_o     = r'{',
  brace_c     = r'}',

  comma       = r',',
  semi        = r';',

  # Order-dependent patterns.
  dot3        = r'\.\.\.',
  dot2        = r'\.\.',
  dot         = r'\.',

  eq3         = r'===',
  eq2         = r'==',
  eq          = r'=',

  exclaim2_eq = r'!==',
  exclaim_eq  = r'!=',
  exclaim     = r'!',

  arrow_r     = r'->',
  dash_eq     = r'-=',
  dash        = r'-',

  plus_eq     = r'\+=',
  plus        = r'\+',

  amp_eq      = r'&=',
  amp2        = r'&&',
  amp         = r'&',

  pipe2       = r'\|\|',
  pipe_eq     = r'\|=',
  pipe        = r'\|',

  at_eq       = r'@=',
  at          = r'@',

  caret_eq    = r'\^=',
  caret       = r'\^',

  colon_eq    = r':=',
  colon       = r':',

  slash2_eq   = r'//=',
  slash2      = r'//',

  slash_eq    = r'/=',
  slash       = r'/',

  star2_eq    = r'\*\*=',
  star2       = r'\*\*',
  star_eq     = r'\*=',
  star        = r'\*',

  qmark2      = r'\?\?',
  qmark       = r'\?',

  shift_l_eq  = r'<<=',
  shift_l     = r'<<',
  le          = r'<=',
  arrow_l     = r'<-',
  lt          = r'<',

  shift_r_eq  = r'>>=',
  shift_r     = r'>>',
  ge          = r'>=',
  gt          = r'>',

  percent_eq  = r'%=',
  percent     = r'%',

  tilde_eq    = r'~=',
  tilde       = r'~',

  backslash   = r'\\',
)



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

import os
from typing import Any, ContextManager, IO, List, Tuple, Union

from .typing import OptBaseExc, OptTraceback, OptTypeBaseExc


File = Union[None, int, IO]


FR, FW, BR, BW = range(4)


class DuplexPipe(ContextManager):

  def __init__(self) -> None:
    self._files: List[File] = list(os.pipe() + os.pipe())

  def __del__(self) -> None:
    for i in range(4): self.close(i)

  def __exit__(self, exc_type:OptTypeBaseExc, exc_value:OptBaseExc, traceback:OptTraceback) -> None:
    for i in range(4): self.close(i)

  def close(self, index:int) -> None:
    f = self._files[index]
    if f is None: return
    elif isinstance(f, int): os.close(f)
    else: f.close()
    self._files[index] = None

  def _fd(self, idx:int) -> int:
    f = self._files[idx]
    if isinstance(f, int): return f
    if f is None: raise Exception('DuplexPipe file has already been closed')
    raise Exception('DuplexPipe file descriptor has already been converted to a file object')

  def _file(self, idx:int, **kwargs:Any) -> IO:
    f = self._files[idx]
    if f is None: raise Exception('DuplexPipe file has already been closed')
    if isinstance(f, int):
      f = open(f, **kwargs)
      self._files[idx] = f
    return f

  @property
  def left_fds(self) -> Tuple[int, int]:
    return (self._fd(BR), self._fd(FW))

  @property
  def right_fds(self) -> Tuple[int, int]:
    return (self._fd(FR), self._fd(BW))

  def left_files(self, **kwargs:Any) -> Tuple[IO, IO]:
    return (self._file(BR, mode='r', **kwargs), self._file(FW, mode='w', **kwargs))

  def right_files(self, **kwargs:Any) -> Tuple[IO, IO]:
    return (self._file(FR, mode='r', **kwargs), self._file(BW, mode='w', **kwargs))

  def close_right(self) -> None:
    for i in (FR, BW): self.close(i)



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from datetime import (date as Date, datetime as DateTime, time as Time, timedelta as TimeDelta, timezone as TimeZone,
  tzinfo as TZInfo)
from typing import TypeVar


sec_per_min = 60
sec_per_hour = sec_per_min * 60
sec_per_day = sec_per_hour * 24

months = ('January', 'February', 'March', 'April', 'May', 'June',
  'July', 'August', 'September', 'October', 'November', 'December')

month_days = {
  1: 31,
  2: 28,
  3: 31,
  4: 30,
  5: 31,
  6: 30,
  7: 31,
  8: 31,
  9: 30,
  10: 31,
  11: 30,
  12: 31,
}


def days_in_year(year:int) -> int:
  if year % 4: return 365
  if year % 100: return 366
  if year % 400: return 365
  return 366


def is_leap_year(year:int) -> bool:
  return not bool((year%4) and (year%100 or not (year%400)))


DateLike = TypeVar('DateLike', Date, DateTime)


class DateDelta:

  def __init__(self, years:int=0, months:int=0) -> None:
    self.years = years
    self.months = months

  def __repr__(self) -> str:
    return f'{type(self).__name__}(years={self.years!r}, months={self.months!r})'

  def __add__(self, other:DateLike) -> DateLike:
    carry_years, month0 = divmod(other.month + self.months - 1, 12)
    year = other.year + self.years + carry_years
    month = month0 + 1
    if month == 2 and other.day == 29 and not is_leap_year(year):
      return other.replace(year=year, month=month, day=28)
    else:
      return other.replace(year=year, month=month)


def parse_datetime(string: str, fmt='%Y-%m-%d %H:%M:%S') -> DateTime:
  return DateTime.strptime(string, fmt)


def parse_date(string: str, fmt='%Y-%m-%d') -> Date:
  return DateTime.strptime(string, fmt).date()


def dt_for_utc_ts(ts:float) -> DateTime:
  return DateTime.fromtimestamp(ts, TimeZone.utc)


def next_day(date_: DateLike) -> DateLike:
  return date_ + TimeDelta(days=1)

def next_week(date_: DateLike) -> DateLike:
  return date_ + TimeDelta(days=7)



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from typing import Any, Callable, ContextManager, Generic, Iterable, List, TextIO, TypeVar

from .io import err_progress, errL, writeL
from .typing import OptBaseExc, OptTraceback, OptTypeBaseExc


_T = TypeVar('_T')
_I = TypeVar('_I')
_O = TypeVar('_O')
Stage = Callable[[_I], _O]
Pred = Callable[[_I], bool]
Put = Callable[[_I], None]


class _DropItem(Exception):
  'Exception signals the Transformer to immediately drop the current item and continue to the next one.'


class Transformer(Generic[_T], ContextManager):
  '''
  A data transformation pipeline.
  '''

  def __init__(self, iterable: Iterable[_T], log_stem:str, log_index_width=2, progress_frequency=0.1) -> None:

    self.iterable = iterable
    self.log_stem = log_stem
    self.log_index_width = log_index_width
    self.progress_frequency = progress_frequency

    self.stages: List[Callable] = []
    self.stage_names: List[str] = []
    self.counts: List[int] = []
    self.log_files: List[TextIO] = []


  def __exit__(self, exc_type:OptTypeBaseExc, exc_value:OptBaseExc, traceback:OptTraceback) -> None:
    if exc_type: return
    self.run()


  def _add_stage(self, name: str, fn: Stage) -> None:
    if name in self.stage_names: # makes adding all stages quadratic time, but len(stages) is low.
      raise KeyError('Transformer: stage name is a duplicate: {}'.format(name))
    self.stage_names.append(name)
    self.stages.append(fn)
    if len(self.counts) != len(self.stages): # this stage has no logger.
      assert len(self.counts) == len(self.stages) - 1
      self.counts.append(-1)


  def _mk_log_fn(self, mode: str, name: str, ext: str) -> Callable[..., None]:
    'create a dedicated log file and logging function for use by a particular stage.'
    index = len(self.stages) # Caller has not added itself as a stage yet.
    path = f'{self.log_stem}{index:0{self.log_index_width}}-{mode}-{name}{ext}'
    f = open(path, 'w')
    self.log_files.append(f)

    counts = self.counts
    assert len(counts) == index
    counts.append(0)

    def log_fn(*items: Any) -> None:
      counts[index] += 1
      writeL(f, *items)

    return log_fn


  def flag(self, pred: Pred) -> None:
    '''
    add a flag stage.
    the decorated function function should take an input item,
    and return a truth value indicating whether the item should be logged.
    '''
    name = pred.__name__
    log_fn = self._mk_log_fn('flag', name, '.txt')

    def flag_fn(item: _I) -> _I:
      if pred(item):
        log_fn(f'{item!r}\n')
      return item

    self._add_stage(name, flag_fn)


  def drop(self, pred: Pred) -> None:
    '''
    add a drop stage.
    the decorated function should be a predicate that takes an input item,
    and returns a truth value indicating whether the item should be dropped from the pipeline.
    drops are logged as deltas.
    '''
    name = pred.__name__
    log_fn = self._mk_log_fn('drop', name, '.diff')

    def drop_fn(item: _I) -> _I:
      if pred(item):
        log_fn(f'- {item!r}\n')
        raise _DropItem
      return item

    self._add_stage(name, drop_fn)


  def keep(self, pred: Pred) -> None:
    '''
    add a keep stage.
    the decorated function should be a predicate that takes an input item,
    and returns a truth value indicating whether the item should continue in the pipeline.
    no logging is performed.
    '''

    def keep_fn(item: _I) -> _I:
      if pred(item):
        return item
      raise _DropItem

    self._add_stage(pred.__name__, keep_fn)


  def edit(self, fn: Stage) -> None:
    '''
    add an edit stage.
    the decorated function should return either the original item or an edited item.
    if the input is not equal to the output then the edit is logged as a delta.
    '''
    name = fn.__name__
    log_fn = self._mk_log_fn('edit', name, '.diff')

    def edit_fn(item: _I) -> _O:
      edited = fn(item)
      if edited != item:
        log_fn(f'- {item!r}\n+ {edited!r}\n')
      return edited # type: ignore

    self._add_stage(name, edit_fn)


  def convert(self, fn: Stage) -> None:
    '''
    add a conversion stage.
    the decorated function should convert the input item to an output item.
    no logging is performed.'
    '''
    self._add_stage(fn.__name__, fn)


  def put(self, fn: Put) -> None:
    '''
    add an output stage.
    the decorated function should output the item, either to stdout or to a file.
    The decorated function does not return the item;
    this is handled by the decorator wrapper.
    no logging is performed.
    '''
    def put_fn(item: _I) -> _I:
      fn(item)
      return item

    self._add_stage(fn.__name__, put_fn)


  def run(self) -> None:

    if not self.stages:
      errL('Transformer: WARNING: no transform functions found; '
        "transformation stage functions must be decorated with the transformer's convert/drop/edit/flag properties.")

    # create local bindings to avoid attribute lookups in the loop.
    stages = self.stages

    errL(f'◊ transform stages: {", ".join(self.stage_names)}.')

    for item in err_progress(self.iterable, 'transform', 'items', frequency=self.progress_frequency):
      try:
        for fn in stages:
          item = fn(item)
      except _DropItem:
        continue

    for f in self.log_files:
      f.close()

    for i, (name, count) in enumerate(zip(self.stage_names, self.counts)):
      c = '-' if count == -1 else count
      width = self.log_index_width
      errL(f'◊   {i:0{width}}-{name}: {c}')




# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from ctypes import CDLL, c_char_p, c_uint, c_int, set_errno
from errno import ENOENT, ENOTSUP
from os import strerror
from os.path import dirname, isdir
from typing import Callable, Optional
import sys


__all__ = ['clone', 'volume_supports_clone']


if sys.platform == 'darwin':
  libSystem = CDLL('libSystem.dylib', use_errno=True)
  clonefile = libSystem.clonefile # (UnsafePointer<Int8>!, UnsafePointer<Int8>!, UInt32) -> Int32
  clonefile.argtypes = (c_char_p, c_char_p, c_uint)
  clonefile.restype = c_int
else:
  def clonefile(src:str, dst:str, flags:int) -> int:
    raise Exception(f'clonefile is not available on this platform: {sys.platform}')


CLONE_NOFOLLOW = 1 # Int32
CLONE_NOOWNERCOPY = 2 # Int32


def clone(src:str, dst:str, follow_symlinks:bool=True, preserve_owner:bool=True, fallback:Optional[Callable[..., None]]=None) -> None:
  '''
  Clone a file using the file system's copy-on-write semantics if available (e.g. APFS); otherwise copy.
  '''
  flags = (0 if follow_symlinks else CLONE_NOFOLLOW) | (0 if preserve_owner else CLONE_NOOWNERCOPY)
  res = clonefile(src.encode(), dst.encode(), flags)
  if res == 0: return
  assert res == -1
  en = set_errno(0)
  assert en != 0, (src, dst, res, en)
  if en == ENOTSUP and fallback is not None: # cloning not supported.
    fallback(src=src, dst=dst, follow_symlinks=follow_symlinks, preserve_owner=preserve_owner)
  elif en == ENOENT: # one of the files or intervening directories does not exist.
    dst_dir = dirname(dst)
    if dst_dir and not isdir(dst_dir): raise NotADirectoryError(dst_dir)
  # TODO: more elaborate diagnosis.
  raise OSError(en, strerror(en), src)


def volume_supports_clone() -> bool:
  '''
  Not all volumes support clonefile().
  A volume can be tested for clonefile() support by using getattrlist(2) to get the volume capabilities attribute ATTR_VOL_CAPABILITIES, and then testing the VOL_CAP_INT_CLONE flag.
  '''
  raise NotImplementedError



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from argparse import ArgumentParser
from dataclasses import dataclass
from typing import Any, Callable, Iterable, List
from tolkien import Source, Token

from ..desc import outD
from ..iterable import iter_values
from ..lex import Lexer
from ..loader import load
from ..parse import Adjacency, Atom, Choice, Left, Parser, Precedence, Struct


def main() -> None:
  '''
  # Syntax:

  ## Predicates
  * `NAME`: a type name.

  ## Filters
  * `*PREDICATE`: search the input object and its children using the specified predicate.
  * `>PREDICATE`: filter the input object using the specified predicate.
  '''

  parser = ArgumentParser(description='Object Query tool.')
  parser.add_argument('paths', nargs='+', help='Path to query.')
  parser.add_argument('-query', nargs='*', default=[], help='Query string.')
  parser.add_argument('-interactive', action='store_true', help='interactive mode.')
  args = parser.parse_args()

  if not args.paths and args.interactive:
    exit(f'interactive mode requires a path argument.')

  roots = [load(p) for p in args.paths]


  if args.interactive:
    interactive_loop(args.query, roots)
  else:
    query = parse_query(' '.join(args.query))
    for result in query.run(roots):
      outD(result)


# Evaluation.

def interactive_loop(query_src:str, roots:List[Any]) -> None:
  exit("INTERACTIVE LOOP NOT YET IMPLEMENTED.")


# Query language.

Predicate = Callable[[Any],bool]


class Query:
  def run(self, stream:Iterable[Any]) -> Iterable[Any]:
    raise NotImplementedError


class PassQuery(Query):
  def run(self, stream:Iterable[Any]) -> Iterable[Any]:
    return stream


@dataclass
class BinaryQuery(Query):
  left:Query
  right:Query


class ChainQuery(BinaryQuery):
  def run(self, stream:Iterable[Any]) -> Iterable[Any]:
    return self.right.run(self.left.run(stream))


class ChildQuery(BinaryQuery):
  def run(self, stream:Iterable[Any]) -> Iterable[Any]:
    return self.right.run(child for result in self.left.run(stream) for child in iter_values(result))


@dataclass
class FilterQuery(Query):
  predicate:Predicate

  def run(self, stream:Iterable[Any]) -> Iterable[Any]:
    p:Predicate = self.predicate # type: ignore # long-standing mypy bug.
    for el in stream:
      if p(el):
        yield el


@dataclass
class SearchQuery(Query):
  predicate:Predicate

  def run(self, stream:Iterable[Any]) -> Iterable[Any]:
    p:Predicate = self.predicate # type: ignore # long-standing mypy bug.
    for el in stream:
      if p(el):
        yield el
      else:
        yield from self.run(iter_values(el))


# Parsing.


def parse_query(src:str) -> Query:
  if not src or src.isspace(): return PassQuery()
  query = parser.parse_or_fail('oq', Source('query', src))
  assert isinstance(query, Query), query
  return query

def mk_type_pred(source:Source, token:Token) -> Predicate:
  type_name = source[token]
  def type_pred(obj:Any) -> bool:
    return type(obj).__name__ == type_name or (isinstance(obj, dict) and obj.get('') == type_name)
  return type_pred



# Lexer.

lexer = Lexer(flags='x', patterns=dict(
  line=r'\n',
  space=r'\s+',
  name=r'[-\w]+',
  par_o=r'\(',
  par_c=r'\)',
  sqb_o=r'\[',
  sqb_c=r'\]',
  cub_o=r'\{',
  cub_c=r'\}',
  star=r'\*',
  dot=r'\.',
  pipe=r'\|',
))

parser = Parser(lexer, dict(
  query=Precedence(
    ('filter', 'search'),
    Left(
      Adjacency(transform=lambda source, token, left, right: ChildQuery(left, right))
    ),
  ),
  search=Choice('pred', transform=lambda source, token, name, predicate: SearchQuery(predicate)),
  filter=Struct('dot', 'pred', transform=lambda source, token, fields: FilterQuery(fields[1])),

  pred=Choice('type_pred', transform=lambda source, token, name, pred: pred),

  type_pred=Atom('name', mk_type_pred),
  ),
  drop=('line', 'space'))



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

'''
Generate and print informative schemas from sets of example object trees.
'''

from typing import Any, Counter, DefaultDict, NamedTuple, Optional, TextIO
from .string import iter_excluding_str


class Schema(NamedTuple):
  '''
  A schema represents the aggregate of values occurring at a structural position in a set of data trees.
  atoms: counts atom values.
  seqs: maps all of the occurring element types to schemas.
  dicts: maps occurring keys to a DefaultDict of occurring value types to schemas.
  All sequence types are lumped together.
  All dict classes are lumped together.
  '''
  atoms: Counter
  seqs: DefaultDict
  dicts: DefaultDict


def _mk_schema() -> Schema:
  return Schema(atoms=Counter(), seqs=DefaultDict(_mk_schema), dicts=DefaultDict(_dd_of_schemas))

def _dd_of_schemas() -> DefaultDict:
  return DefaultDict(_mk_schema)


def _compile_schema(node: Any, schema: Schema):
  if isinstance(node, dict):
    # dict schemas have two layers: node key and node val type.
    for k, v in node.items():
      _compile_schema(v, schema.dicts[k][type(v)])
  else:
    try:
      it = iter_excluding_str(node) # will raise TypeError if str or not iterable.
    except TypeError: # n is an atom.
      schema.atoms[node] += 1
    else: # n is iterable.
      # iterable schemas have one layer: the node element type.
      for el in it:
        _compile_schema(el, schema.seqs[type(el)])


def compile_schema(*nodes: Any, schema: Schema=None) -> Schema:
  '''
  Generate or update a `Schema` from one or more example objects.
  Each object (JSON or similar generic collections) is explored
  and type information about constituent dictionary, sequence, and atom values is saved.

  Each node of the schema represents a level of the aggregate structure.
  `Schema` objects consist of:
  * atoms: a Counter of all atoms (non-Dict, non-sequence) "leaf" values.
  * seqs: a mapping from types to element schemas.
  * dicts: a two-level mapping from keys to types to value schemas.
  '''
  if schema is None:
    schema = _mk_schema()
  for node in nodes:
    _compile_schema(node, schema)
  return schema


def _unique_el(counter: Counter) -> Any:
  'Return the first element of the counter whose count is 1.'
  for k, c in counter.items():
    if c == 1: return k
  raise ValueError(counter)


def _write_schema(file: TextIO, schema: Schema, count_atoms: bool, inline: bool, indent: str, root: bool) -> None:
  '''
  Note: _write_schema expects its caller to not have emitted a trailing newline.
  This allows it to decide whether or not to inline monomorphic type information.
  '''

  def put(*items: Any):
    print(*items, sep='', end='', file=file)

  def put_types(prefix: str, symbol: str, subindent: str, types: dict):
    for t, subschema, in sorted(types.items(), key=lambda item: item[0].__name__):
      put(prefix, symbol, t.__name__)
      _write_schema(file, subschema, count_atoms=count_atoms, inline=inline, indent=subindent, root=False)

  if not any(schema): # should only happen for root; other schemas are created on demand.
    put(indent, 'empty')
    return
  if count_atoms and schema.atoms:
    repeated_atoms = sorted(((c, v) for v, c in schema.atoms.items() if c > 1), reverse=True)
    unique_count = len(schema.atoms) - len(repeated_atoms)
    for c, v in repeated_atoms:
      put(indent, '#', c, ' ', repr(v))
    if unique_count > 1:
      put(indent, '+', unique_count)
    elif unique_count == 1: # find the unique element.
      put(indent, '#1 ', _unique_el(schema.atoms))
  if schema.seqs:
    if inline and len(schema.seqs) == 1 and not schema.atoms and not schema.dicts:
      prefix = ('' if root else ' ')
    else:
      prefix = indent
    put_types(prefix=prefix, symbol='* ', subindent=(indent + '| '), types=schema.seqs)
  if schema.dicts:
    for k, types in sorted(schema.dicts.items()):
      put(indent, repr(k))
      # Inlining for dictionaries is simpler, because we can always inline after the key we just emitted.
      prefix = ' ' if (inline and len(types) == 1) else indent
      put_types(prefix=prefix, symbol=': ', subindent=(indent + '. '), types=types)


def write_schema(file: TextIO, schema: Optional[Schema]=None, count_atoms=False, inline=True, indent='', end='\n') -> None:
  '''
  Write `schema` to file `file`.
  If `count_atoms` is true, then histograms of atom values are emitted.
  If `inline` is false, then monomorphic type names are never inlined,
  resulting in longer but more regular output.
  '''
  if schema is None:
    file.write(indent + 'empty')
  else:
    _write_schema(file, schema=schema, count_atoms=count_atoms, inline=inline, indent='\n' + indent, root=True)
  file.write(end)




# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from typing import Iterable


class ParseError(Exception): pass

def pat_dependency(src_path:str, src_lines:Iterable[str]) -> str:
  '''
  Return a list of dependencies.
  A .pat file always has a single dependency: the source file it patches.
  '''
  it = iter(src_lines)
  version_line = next(it, '')
  if not version_line.startswith('pat v'):
    raise ParseError(f'pat error: {src_path}:1:1: malformed version string.')
  orig_line = next(it, '')
  orig_path = orig_line.strip()
  if not orig_path:
    raise ParseError(f'pat error: {src_path}:2:1: line specifying original path is missing or empty.')
  return orig_path



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from argparse import ArgumentParser
from functools import partial
from typing import cast

from ..http.server import HTTPRequestHandler, HTTPServer
from ..task import run


def main() -> None:

  parser = ArgumentParser(description='Serve files from a directory.')
  parser.add_argument('root', default='.', nargs='?', help='Root directory to serve from')
  parser.add_argument('-browse', action='store_true', help='Launch the default system browser')
  parser.add_argument('-chrome',  action='store_true', help='Launch Google Chrome')
  parser.add_argument('-firefox', action='store_true', help='Launch Firefox')
  parser.add_argument('-safari',  action='store_true', help='Launch Safari')
  parser.add_argument('-stp',  action='store_true', help='Launch Safari Technology Preview')

  args = parser.parse_args()
  root = args.root
  address = ('localhost', 8000) # TODO: argparse option.
  host, port = address
  addr_str = f'http://{host}:{port}'
  print(addr_str)

  ignored_paths = {
    'apple-touch-icon-precomposed.png',
  }

  server = HTTPServer(address, cast(type, partial(HTTPRequestHandler, directory=root)))

  # note: the way we tell the OS to open the URL in the browser is a rather suspicious hack:
  # the `open` command returns and then we launch the web server,
  # relying on the fact that together the OS and the launched browser take more time to initiate the request
  # after the `open` process completes than the server does to initialize.
  if args.browse:   run(['open', addr_str])
  if args.chrome:   run(['open', '-a', 'google chrome', addr_str])
  if args.firefox:  run(['open', '-a', 'firefox',       addr_str])
  if args.safari:   run(['open', '-a', 'safari',        addr_str])
  if args.stp:      run(['open', '-a', 'safari technology preview', addr_str])

  try: server.serve_forever()
  except KeyboardInterrupt:
    print('\nKeyboard interrupt received; shutting down server.')
    exit()



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from dataclasses import asdict, is_dataclass
from functools import singledispatch
from typing import Any, Callable, FrozenSet, Set, Type

from .util import memoize


EncodeObj = Callable[[Any],Any]

@singledispatch
def encode_obj(obj:Any) -> Any:
  '''
  Encode an object for serialization, e.g. to json or msgpack.
  This function is used for the default `default` converter by `render_json`, `write_json`, `write_msgpack`.

  Note: it is not possible to encode namedtuple as a JSON dict using a `default` function such as this,
  because the namedtuple gets converted to a list without ever calling `default`.
  '''

  try: it = iter(obj) # Try to convert to a sequence first.
  except TypeError: pass
  else: return list(it)

  if is_dataclass(obj): return asdict(obj)

  if hasattr(obj, '__slots__'):
    slots = all_slots(type(obj))
    slots = slots.union(getattr(obj, '__dict__', ())) # Slots classes may also have backing dicts.
    return {a: getattr(obj, a) for a in slots}

  try: d = obj.__dict__ # Treat other classes as dicts by default.
  except AttributeError: pass
  else:
    if any(k.startswith('_') for k in d): # Only create a new dictionary if necessary.
      return {k:v for k,v in d.items() if not k.startswith('_')}
    else:
      return d

  return str(obj) # convert to string as last resort.


@encode_obj.register
def _(obj:None) -> Any: return obj

@encode_obj.register # type: ignore
def _(obj:bool) -> Any: return obj

@encode_obj.register # type: ignore
def _(obj:int) -> Any: return obj

@encode_obj.register # type: ignore
def _(obj:str) -> Any: return obj

@encode_obj.register # type: ignore
def _(obj:type) -> Any: return repr(obj)


@memoize()
def all_slots(type: Type) -> FrozenSet[str]:
  '''
  Subclasses of slots classes may define their own slots,
  which hold just the additions to the parent class.
  Therefore we need to iterate over the inheritance chain to get all slot names.
  We use __mro__ here, and hope for the best regarding multiple inheritance.
  '''
  slots: Set[str] = set()
  for t in type.__mro__:
    try: s = t.__slots__
    except AttributeError: break
    else:
      if isinstance(s, str): slots.add(s) # single slot.
      else:
        slots.update(s)
  return frozenset(slots)



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from argparse import ArgumentParser
from base64 import b64encode


def main() -> None:
  arg_parser = ArgumentParser('Convert images.')
  arg_parser.add_argument('paths', nargs='+', help='paths to image files.')
  arg_parser.add_argument('-base64', action='store_true', help='output base64 encoding to stdout.')
  args = arg_parser.parse_args()

  for path in args.paths:
    conv_img(path, base64=args.base64)


def conv_img(path:str, base64:bool) -> None:
  b = open(path, 'rb').read()
  if base64:
    encoded = b64encode(b, altchars=None) # type: ignore
    print(f'{path}:\ndata:image/png;base64,{encoded.decode()}')
  else:
    exit('error: `-base64` is currently the only supported option.')

if __name__ == '__main__': main()



# Dedicated to the public domain under CC0: https://creativecommons.org/publicdomain/zero/1.0/.

from enum import Enum
from typing import TypeVar, Union


_T = TypeVar('_T')


class Default(Enum):
  '''
  Singleton class and value to indicate a default parameter value,
  for cases where None is a meaningful user provided value.
  For example: `def f(x:Union[int, Default]=Default._): ...`

  see: https://www.python.org/dev/peps/pep-0484/#support-for-singleton-types-in-unions
  '''
  _ = 0


class Raise(Enum):
  '''
  Raise serves the same purpose as Default,
  but suggests that the default behavior is to raise an exception.
  '''
  _ = 0


RaiseOr = Union[Raise, _T]
